%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (CHEM-E7225) ADVANCED PROCESS CONTROL D
%   #01 - Numerical integration of dynamical systems
%
%   CasADi is a tool to quickly implement and efficiently solve dynamic 
%   optimisation (optimal control) problems. You can also think of it as 
%   a general toolkit for gradient-based optimisation
%
%   CasADI is open-source, LGPL-licensed
%   Download from https://web.casadi.org  (current version: 3.6.4)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear; close all; clc

addpath('~/Documents/MATLAB/casadi')     % <~ Change to the path of the CasADi folder in your computer
import casadi.*

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define the initial value problem (IVP)

% Simulation parameters
Tf = 25;                    % Terminal time (in time units)
K  = 2e4;                   % Number of time-nodes (dimensionless)
dT = Tf / K;                % Timestep length (\Delta t)
t = (0:K-1)*dT;             % Time-nodes (in time units)

% Model dynamics and parameters parameters
f = @Brusselator;           % This creates a "alias" for the state-equation function

A = 1; B = 1.7;             % Parameters of the system (individually)
px = [A, B];                % Vector of parameters (i.e., \theta in the slides)

Nx = 2; Nu = 0; Np = 2;     % Dimensions of the system

% Initial state
x0  = [3; 4];

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ODE15s
%  User friendly interface to integrate ODEs/DAEs of type xplus = f(t,x,u|px) 
%  from initial conditon x(0), over [0,Tf], subjected to controls u(t)

tic
[~, X_ode15s] = ode15s(@(t,x) f(t,x,[],px), t, x0);

toc_ode15s = toc;
disp( "Execution time (ODE15s): " + toc_ode15s + "s");

plotSimulation(t, X_ode15s', 'ODE15s')

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Explicit Euler
%  Numerical method to solve IVPs given xplus = f(t,x,u|p) and initial x(0),
%  over [0,Tf], by sequentially solving the formula
%
%       x_k+1 = x_k + f(k*dT,x_k,u_k|p)*dT
%
%  for k = [0,1,...,Tf/dT] and x_0 = x(0), given a step-size dT > 0.

X_ee = nan(Nx,K); 
X_ee(:,1) = x0;

tic
for k = 1:K-1
    X_ee(:,k+1) = X_ee(:,k) + f([], X_ee(:,k), [], px) * dT;
end

toc_ee = toc;
disp( "Execution time (EE): " + toc_ee + "s");

plotSimulation(t, X_ee, 'Explict Euler')

    
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Explicit Range-Kutta Order 4 (RK4)
%  Numerical method to solve IVPs given xplus = f(t,x,u|p) and initial x(0),
%  over [0,Tf], by sequentially solving the formulas
%
%       k1 = f(k*dT     , x_k        , u_k|p)
%       k2 = f(k*dT+dT/2, x_k+dT*k1/2, u_k|p)
%       k3 = f(k*dT+dT/2, x_k+dT*k2/2, u_k|p)
%       k4 = f(k*dT+dT  , x_k+dT*k3  , u_k|p)
%
%       x_k+1 = x_k + dT/6*(k1 + 2*k2 + 2*k3 + k4)
%
%  for k = [0,1,...,Tf/dT] and x_0 = x(0), given a step-size dT.

X_rk4 = nan(Nx,K); 
X_rk4(:,1) = x0;

tic
for k = 1:K-1
    k1 = f([], X_rk4(:,k),               [], px);
    k2 = f([], X_rk4(:,k) + k1 * dT / 2, [], px);
    k3 = f([], X_rk4(:,k) + k2 * dT / 2, [], px);
    k4 = f([], X_rk4(:,k) + k3 * dT,     [], px);

    X_rk4(:,k+1) = X_rk4(:,k) + dT/6 * (k1 + 2*k2 + 2*k3 + k4);
end

toc_rk4 = toc;
disp( "Execution time (RK4): " + toc_rk4 + "s");

plotSimulation(t, X_rk4, 'Range-Kutta 4th Order')


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CasADi (using the Integrator function)
%  An integrator is a function F: (R+ x R^n x R^p) -> R^n;  F(t_0,x(0),p) = x(t_f)
%
%  Define an integrator using an ODE symbolic expression 
%     x_k+1 = F_ode(t,x|p_ode), 
%  with the parameters p_ode = [u, p].

% First, we create an integrator F_casadi with a chosen method ...
X = MX.sym('x', Nx);
U = MX.sym('u', Nu);
P = MX.sym('px',Np);

ODE = struct('x',X, 'p',[U;P], 'ode',f([], X, [], P));   % The ODE object for the integrator

F_method = 'cvodes';    % Options: {'rk', 'idas', 'cvodes', 'collocation'}
F_casadi = integrator('F', F_method, ODE, struct('tf',dT));

% ... then we use the Integrator on the simulation loop
X_casadi = nan(Nx,K); 
X_casadi(:,1) = x0;

tic
for k = 1:K-1
    F_casadi_res = F_casadi('x0', X_casadi(:,k), 'p', [; px]);
    X_casadi(:,k+1) = full( F_casadi_res.xf' );
end

toc_casadi = toc;
disp( "Execution time (CasADi, w/ "+F_method+"): " + toc_casadi + "s");

plotSimulation(t, X_casadi, "CasADi ("+F_method+")")

% --